/*
    objreader: .obj file reader's source file
    Description: directly creates a display list from given file
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "objfile.h"

/* Structure for our material library */
struct TMaterial
{
    char materialName[256];
    float ambientMat[4];
    float diffuseMat[4];
    float specularMat[4];
    int shininess;
};

struct TMaterial *MaterialLib = NULL;
unsigned int NumOfMaterial;
Point3D MinBBox;
Point3D MaxBBox;

/* Structure for all vertex data encountered in .obj file */
struct TObjectNode
{
    unsigned int NumOfVertices;
    unsigned int NumOfNormals;
    unsigned int NumOfTextures;
    Point3D *Vertex;
    Point3D *VertexNormal;
    Point3D *VertexTexture;
};

/* Enumerating whatever... Internal use only */
/* PS: Not all keyword has been enumerated here... */
enum
{
    DATA = 0,                   /* The data, numerical data */
    KEYWORD_UNKNOWN,            /* Unknown a keyword */
    KEYWORD_V,                  /* Vertex */
    KEYWORD_VT,                 /* Texture coordinate */
    KEYWORD_VN,                 /* Normal vector */
    KEYWORD_F,                  /* Face */
    KEYWORD_USEMTL,             /* Material */
    KEYWORD_END,                /* End */
    VERT_READ_OK,               /* Read vertex OK */
    FACE_READ_OK,               /* Read face OK */
    ERROR_READ_VERT,            /* Error in ReadVertex function */
    ERROR_READ_FACE,            /* Error on ReadFace function */
    KEYWORD_COMMENT             /* Comment line */
};

/*
    FUNCTION PROTOTYPE (These functions are not accessible from outside)
*/

/* Count the vertices & faces to determine how much memory we should allocate */
void CountVertices(FILE *fileptr, unsigned int *v, unsigned int *vt,
    unsigned int *vn);
unsigned int CountFaces(FILE *fileptr);

/* Read all vertices data */
int ReadVerticesData(FILE *fileptr, struct TObjectNode *objnode);
/* ReadVertex will return next keyword encountered */
int ReadVertex(FILE *fileptr, Point3D *vertex);
/* Read all faces data */
int ReadFacesData(FILE *fileptr, struct TObjectNode objnode, int numFaces);
/* ReadFace will return error status if encountering error */
int ReadFace(char *fileptr, struct TObjectNode objnode);

/* Parse keyword here, in this function, add everything you need here */
int ParseKeyword(char *string);

/*
    THE IMPLEMENTATION
*/

/*
    ReadMaterialLibrary()
    The function accessible to outside world.
    This will initialize the material library.
    Just pass the file name of your material library.
*/
int ReadMaterialLibrary(const char *material)
{
    FILE *input_file = NULL;
    char buffer1[64], buffer2[64];
    int i;
    
    /* Open material library file */
    if((input_file = fopen(material, "r")) == NULL)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot read %s file\n", material);
        #endif
        return(0);        
    }
    
    /* How many material here */
    fscanf(input_file, "%[^:]:%[^\n]", buffer1, buffer2);
    NumOfMaterial = (unsigned int) atoi(buffer2);
    #ifdef _DEBUG_
        fprintf(stderr, "NumOfMaterial = %d\n", NumOfMaterial);
    #endif
    
    /* If there's any previous material library, free it */
    if(MaterialLib)
    {
        free(MaterialLib);
        MaterialLib = NULL;
    }
    
    /* Allocate memory and check it */
    MaterialLib = (struct TMaterial *) malloc(sizeof(struct TMaterial) *
        NumOfMaterial);
    if(!MaterialLib)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot allocate memory for material\n");
        #endif
        return(0);
    }
    
    /* Get materials */
    for(i = 0; (unsigned int) i < NumOfMaterial; i++)
    {
        if(feof(input_file) && (unsigned int) i < NumOfMaterial)
        {
            #ifdef _DEBUG_
                fprintf(stderr, "Error: Incomplete material\n");
            #endif
            return(0);            
        }
        fscanf(input_file, "%s %f %f %f %f %f %f %f %f %f %f %f %f %d\n",
            MaterialLib[i].materialName,
            &MaterialLib[i].ambientMat[0],
            &MaterialLib[i].ambientMat[1],
            &MaterialLib[i].ambientMat[2], 
            &MaterialLib[i].ambientMat[3], 
            &MaterialLib[i].diffuseMat[0],
            &MaterialLib[i].diffuseMat[1],
            &MaterialLib[i].diffuseMat[2], 
            &MaterialLib[i].diffuseMat[3], 
            &MaterialLib[i].specularMat[0],
            &MaterialLib[i].specularMat[1],
            &MaterialLib[i].specularMat[2], 
            &MaterialLib[i].specularMat[3], 
            &MaterialLib[i].shininess);
    }
    return(1);
}

/*
    ReadObject()
    The function accessible to outside world.
    It will automatically read file [filename].
    Result: the display list is initialized.
    Return value: display list's number.
*/
GLuint ReadObject(const char *filename, Object3D *object)
{
    FILE *input_file = NULL;
    struct TObjectNode theObject;
    int numOfFaces;
    GLuint displayList;

    /* First... Open it... */
    if((input_file = fopen(filename, "r")) == NULL)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot read %s file\n", filename);
        #endif
        return(0);
    }
    
    /* Count all vertices */
    CountVertices(input_file, &theObject.NumOfVertices,
        &theObject.NumOfTextures, &theObject.NumOfNormals);
    /* And faces */
    numOfFaces = CountFaces(input_file);
    
    /* Allocate memory for all vertices data */
    theObject.Vertex = (Point3D *) malloc(sizeof(Point3D) *
        theObject.NumOfVertices);
    theObject.VertexNormal = (Point3D *) malloc(sizeof(Point3D) *
        theObject.NumOfNormals);
    theObject.VertexTexture = (Point3D *) malloc(sizeof(Point3D) *
        theObject.NumOfTextures);
        
    if(!theObject.Vertex || !theObject.VertexNormal ||
        !theObject.VertexTexture)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot allocate memory for vertex data\n");
        #endif
        return(0);
    }
    
    displayList = glGenLists(1);
    glNewList(displayList, GL_COMPILE);
    if(displayList == 0)
    {
        #ifdef _DEBUG_
            fprintf(stderr, "Error: Cannot generate OpenGL list\n");
        #endif
        return(0);
    }

    #ifdef _DEBUG_
        fprintf(stderr, "Number of vertices = %d\n",
            theObject.NumOfVertices);
        fprintf(stderr, "Number of normals = %d\n",
            theObject.NumOfNormals);
        fprintf(stderr, "Number of texture coordinates = %d\n",
            theObject.NumOfTextures);
        fprintf(stderr, "Number of faces = %d\n",
            numOfFaces);
        fprintf(stderr, "Display List Number = %d\n", displayList);
    #endif
    
    /* Initialize bounding box */
    MinBBox.x = MinBBox.y = MinBBox.z = 99999.0f;
    MaxBBox.x = MaxBBox.y = MaxBBox.z = -99999.0f;
    /* Read vertices here... */
    if(ReadVerticesData(input_file, &theObject) != VERT_READ_OK)
    {
        fclose(input_file);
        return(0);
    }
    
    /* Read faces here... */
    if(ReadFacesData(input_file, theObject, numOfFaces) != FACE_READ_OK)
    {
        fclose(input_file);
        return(0);
    }
    
    glEndList();

    fclose(input_file);
    
    free(theObject.Vertex);
    free(theObject.VertexNormal);
    free(theObject.VertexTexture);
    
    object->DisplayListNum = displayList;
    object->MinimumBoundingBox.x = MinBBox.x;
    object->MinimumBoundingBox.y = MinBBox.y;
    object->MinimumBoundingBox.z = MinBBox.z;
    object->MaximumBoundingBox.x = MaxBBox.x;
    object->MaximumBoundingBox.y = MaxBBox.y;
    object->MaximumBoundingBox.z = MaxBBox.z;
    
    return(displayList);
}

/*
    CountVertices()
    Counts all vertex related data (keyword v, vt, vn).
    Reads data from [fileptr], give all the output as v, vt, vn.
*/
void CountVertices(FILE *fileptr, unsigned int *v, unsigned int *vt,
    unsigned int *vn)
{
    unsigned int countv = 0;
    unsigned int countvt = 0;
    unsigned int countvn = 0;
    char buffer[64];
    
    while(!feof(fileptr))
    {
        fscanf(fileptr, "%s\n", buffer);
        if(buffer[0] == '#')
            fscanf(fileptr, "%[^\n]\n", buffer);
        else if(buffer[0] == 'v' && buffer[1] == 't')
            countvt++;
        else if(buffer[0] == 'v' && buffer[1] == 'n')
            countvn++;
        else if(buffer[0] == 'v')
            countv++;
    }
    
    *v = countv;
    *vt = countvt;
    *vn = countvn;

    rewind(fileptr);
}

/*
    CountFaces()
    Counts all face data appeared in [fileptr].
    Returns number of faces.
*/
unsigned int CountFaces(FILE *fileptr)
{
    unsigned int count = 0;
    char buffer[64];
    
    while(!feof(fileptr))
    {
        fscanf(fileptr, "%s\n", buffer);
        if(buffer[0] == '#')
            fscanf(fileptr, "%[^\n]\n", buffer);
        else if(buffer[0] == 'f' && buffer[1] == 0x00)
            count++;
    }

    rewind(fileptr);
    return(count);
}

/*
    ReadVerticesData()
    Parses all vertex related data (keyword v, vt, vn).
    Reads from [fileptr] to initialize [objnode]. Returns read error status.
*/
int ReadVerticesData(FILE *fileptr, struct TObjectNode *objnode)
{
    char buffer[64];
    unsigned int vertexcnt = 0, normalcnt = 0, texturecnt = 0;
    int key;
    
    /* Parse the first keyword */
    fscanf(fileptr, "%s\n", buffer);
    key = ParseKeyword(buffer);
    while(vertexcnt < objnode->NumOfVertices ||
        normalcnt < objnode->NumOfNormals ||
        texturecnt < objnode->NumOfTextures)
    {
        switch(key)
        {
        case KEYWORD_COMMENT: /* If it is a comment */
            fscanf(fileptr, "%[^\n]\n", buffer);
            fscanf(fileptr, "%s\n", buffer);
            key = ParseKeyword(buffer);
            break;
        case KEYWORD_V: /* If it is vertex... */
            key = ReadVertex(fileptr, &objnode->Vertex[vertexcnt]);
            if(key == ERROR_READ_VERT)
                return(key);
            #ifdef _NUS_DEBUG_
                fprintf(stderr, "vertexcnt = %d\n", vertexcnt);
                fprintf(stderr, "0x%x:\n", &objnode->Vertex[vertexcnt]);
                fprintf(stderr, "-- 0x%x: %f\n", &objnode->Vertex[vertexcnt].x,
                    objnode->Vertex[vertexcnt].x);
                fprintf(stderr, "-- 0x%x: %f\n", &objnode->Vertex[vertexcnt].y,
                    objnode->Vertex[vertexcnt].y);
                fprintf(stderr, "-- 0x%x: %f\n", &objnode->Vertex[vertexcnt].z,
                    objnode->Vertex[vertexcnt].z);
            #endif
            
            /* Calculate bounding box */
            if(objnode->Vertex[vertexcnt].x < MinBBox.x)
                MinBBox.x = objnode->Vertex[vertexcnt].x;
            if(objnode->Vertex[vertexcnt].y < MinBBox.y)
                MinBBox.y = objnode->Vertex[vertexcnt].y;
            if(objnode->Vertex[vertexcnt].z < MinBBox.z)
                MinBBox.z = objnode->Vertex[vertexcnt].z;

            if(objnode->Vertex[vertexcnt].x > MaxBBox.x)
                MaxBBox.x = objnode->Vertex[vertexcnt].x;
            if(objnode->Vertex[vertexcnt].y > MaxBBox.y)
                MaxBBox.y = objnode->Vertex[vertexcnt].y;
            if(objnode->Vertex[vertexcnt].z > MaxBBox.z)
                MaxBBox.z = objnode->Vertex[vertexcnt].z;
            
            vertexcnt++;
            break;
        case KEYWORD_VN: /* If it is normal vector... */
            key = ReadVertex(fileptr, &objnode->VertexNormal[normalcnt]);
            if(key == ERROR_READ_VERT)
                return(key);
            #ifdef _NUS_DEBUG_
                fprintf(stderr, "normalcnt = %d\n", normalcnt);
                fprintf(stderr, "0x%x:\n", &objnode->VertexNormal[normalcnt]);
                fprintf(stderr, "-- 0x%x: %f\n",
                    &objnode->VertexNormal[normalcnt].x,
                    objnode->VertexNormal[normalcnt].x);
                fprintf(stderr, "-- 0x%x: %f\n",
                    &objnode->VertexNormal[normalcnt].y,
                    objnode->VertexNormal[normalcnt].y);
                fprintf(stderr, "-- 0x%x: %f\n",
                    &objnode->VertexNormal[normalcnt].z,
                    objnode->VertexNormal[normalcnt].z);
            #endif
            normalcnt++;
            break;
        case KEYWORD_VT: /* If it is texture coordinate */
            key = ReadVertex(fileptr, &objnode->VertexTexture[texturecnt]);
            if(key == ERROR_READ_VERT)
                return(key);
            #ifdef _NUS_DEBUG_
                fprintf(stderr, "texturecnt = %d\n", texturecnt);
                fprintf(stderr, "0x%x:\n", &objnode->VertexTexture[texturecnt]);
                fprintf(stderr, "-- 0x%x: %f\n",
                    &objnode->VertexTexture[texturecnt].x,
                    objnode->VertexTexture[texturecnt].x);
                fprintf(stderr, "-- 0x%x: %f\n",
                    &objnode->VertexTexture[texturecnt].y,
                    objnode->VertexTexture[texturecnt].y);
            #endif
            texturecnt++;
            break;
        default:
            if(key != KEYWORD_END)
                fscanf(fileptr, "%[^\n]\n", buffer);
            fscanf(fileptr, "%s\n", buffer);
            key = ParseKeyword(buffer);
        }
    }

    /* Always restore pointer position to the beginning of the file */
    rewind(fileptr);
    return(VERT_READ_OK);
}

/*
    ReadVertex()
    Read data from vertex declaration, maximum triplet.
    Input stream is as [fileptr], result will be stored in [vertex].
    Return value is the next keyword found in the stream.
*/
int ReadVertex(FILE *fileptr, Point3D *vertex)
{
    int key;
    char buffer[64];
    short count = 0;
    
    /* Read the first data */
    fscanf(fileptr, "%s\n", buffer);
    key = ParseKeyword(buffer);
    while(key == DATA) /* Read & store data as vertex */
    {
        switch(count) /* Switch to fill appropriate data */
        {
        case 0:
            vertex->x = atof(buffer);
            count++;
            break;
        case 1:
            vertex->y = atof(buffer);
            count++;
            break;
        case 2:
            vertex->z = atof(buffer);
            count++;
            break;
        default: /* Assumption: we only receive triplets, maximum */
            return(ERROR_READ_VERT);
        }
        /* Read next data */
        fscanf(fileptr, "%s\n", buffer);
        key = ParseKeyword(buffer);
    }
    
    return(key);
}

/*
    ReadFacesData()
    Reads faces from stream [fileptr]. Builds 3D object from faces
    using data in [objnode].
    Return value is read error status.
*/
int ReadFacesData(FILE *fileptr, struct TObjectNode objnode, int numFaces)
{
    char buffer[256];
    int count = 0;
    int keyword;
    int currmtl;   /* Indicate current material for faces */

    /* Get the first keyword */
    fscanf(fileptr, "%s\n", buffer);
    keyword = ParseKeyword(buffer);
    while(count < numFaces) /* Just stop if the we got all the faces */
    {
        switch(keyword)
        {
        case KEYWORD_COMMENT:
            fscanf(fileptr, "%[^\n]\n", buffer);
            fscanf(fileptr, "%s\n", buffer);
            keyword = ParseKeyword(buffer);
            break;
        case KEYWORD_USEMTL: /* Set the current material for these faces */
            fscanf(fileptr, "%[^\n]\n", buffer);
            /* Material can be called by using material name */
            for(currmtl = 0; (unsigned int) currmtl < NumOfMaterial; currmtl++)
                if(strcmp(buffer, MaterialLib[currmtl].materialName) == 0)
                {
                    glMaterialfv(GL_FRONT, GL_AMBIENT,
                        MaterialLib[currmtl].ambientMat);
                    glMaterialfv(GL_FRONT, GL_DIFFUSE,
                        MaterialLib[currmtl].diffuseMat);
                    glMaterialfv(GL_FRONT, GL_SPECULAR,
                        MaterialLib[currmtl].specularMat);
                    glMateriali(GL_FRONT, GL_SHININESS,
                        MaterialLib[currmtl].shininess);
                    #ifdef _DEBUG_
                        fprintf(stderr, "Changing material to %d (%s)\n",
                            currmtl, buffer);
                    #endif
                    break;
                }
            fscanf(fileptr, "%s\n", buffer);
            keyword = ParseKeyword(buffer);
            break;
        case KEYWORD_F: /* If it's the face information, do this... */
            fscanf(fileptr, "%[^\n]\n", buffer);
            if(ReadFace(buffer, objnode) == ERROR_READ_FACE)
                return(ERROR_READ_FACE);
            fscanf(fileptr, "%s\n", buffer);
            keyword = ParseKeyword(buffer);
            count++;
            break;
        default:
            if(keyword != KEYWORD_END)
                fscanf(fileptr, "%[^\n]\n", buffer);
            fscanf(fileptr, "%s\n", buffer);
            keyword = ParseKeyword(buffer);
        }
    }
    return(FACE_READ_OK);
}

/*
    ReadFace()
    Reads data for a face in [string], result will be the display list.
    As always, vertex data are stored in [objnode].
    Return value is the error status.
*/
int ReadFace(char *string, struct TObjectNode objnode)
{
    char buffer[64];
    int count = 0, datacnt = 0, entitycnt = 0, bufferptr = 0;
    int vertexnum = 0; 
    int vertexgrp[3];
    int i;

    /* Count the vertices in this face */
    do
    {
        if(string[count] == 0x20)
            datacnt++;
        count++;
        if(string[count] == 0x00)
            datacnt++;            
    }
    while(string[count] != 0x00);
    
    count = 0;
    
    vertexgrp[0] = vertexgrp[1] = vertexgrp[2] = 0;

    switch(datacnt)
    {
    case 1:
        glBegin(GL_POINTS);
        break;
    case 2:
        glBegin(GL_LINES);
        break;
    case 3:
        glBegin(GL_TRIANGLES);
        break;
    case 4:
        glBegin(GL_QUADS);
        break;
    default:
        glBegin(GL_POLYGON);
    }
    
    while(vertexnum < datacnt)
    {
        switch(string[count])
        {
        case 0x00: /* This is the terminator of this data */
        case 0x20: /* We're reaching the beginning of next vertex */
            buffer[bufferptr] = 0x00;
            vertexgrp[entitycnt] = atoi(buffer) - 1;
            /* Describe normal & tex. coord. first, before declaring vertex */
            for(i = 2; i >= 0; i--)
            {
                switch(i)
                {
                case 0:
                    glVertex3f(objnode.Vertex[vertexgrp[i]].x,
                        objnode.Vertex[vertexgrp[i]].y,
                        objnode.Vertex[vertexgrp[i]].z);
                    break;
                case 1:
                    glTexCoord2f(objnode.VertexTexture[vertexgrp[i]].x,
                        objnode.VertexTexture[vertexgrp[i]].y);
                    break;
                case 2:
                    glNormal3f(objnode.VertexNormal[vertexgrp[i]].x,
                        objnode.VertexNormal[vertexgrp[i]].y,
                        objnode.VertexNormal[vertexgrp[i]].z);
               }
            }
            vertexgrp[0] = vertexgrp[1] = vertexgrp[2] = 0;
            entitycnt = 0;
            bufferptr = 0;
            vertexnum++;
            break;
        case '/': /* The data is seperated by slash */
            buffer[bufferptr] = 0x00;
            if(entitycnt >= 3)
                return(ERROR_READ_FACE);
            vertexgrp[entitycnt] = atoi(buffer) - 1;
            bufferptr = 0;
            entitycnt++;
            break;
        default:
            buffer[bufferptr] = string[count];
            bufferptr++;
        }
        count++;
    }

    glEnd();

    return(FACE_READ_OK);
}

/*
    ParseKeyword()
    Parses keyword from string [string] and return keyword code.
*/
int ParseKeyword(char *string)
{
    /* Refer to enumeration on the beginning of the file */
    if(string[0] == '0' || string[0] == '1' || string[0] == '2' ||
        string[0] == '3' || string[0] == '4' || string[0] == '5' ||
        string[0] == '6' || string[0] == '7' || string[0] == '8' ||
        string[0] == '9' || string[0] == '-')
        return(DATA);
    if(string[0] == '#')
        return(KEYWORD_COMMENT);
    if(strcmp(string, "v") == 0)
        return(KEYWORD_V);
    if(strcmp(string, "vt") == 0)
        return(KEYWORD_VT);
    if(strcmp(string, "vn") == 0)
        return(KEYWORD_VN);
    if(strcmp(string, "f") == 0)
        return(KEYWORD_F);
    if(strcmp(string, "usemtl") == 0)
        return(KEYWORD_USEMTL);
    if(strcmp(string, "end") == 0)
        return(KEYWORD_END);
    
    return(KEYWORD_UNKNOWN);
}
